
/******************************************************************************

Copyright (c) 1997, Andrew J. Weber, Jason P. Weber

Copyright information should have been included with this distribution
and may be found at ../meta/COPYRIGHT or obtained via email from weber@acm.org

******************************************************************************/

#include <bld/rule.h>
#include <bld/interp.h> // For BLD_CONTEXT

/***********************************************************************
class BLD_RuleBase
***********************************************************************/
GPL_Boolean BLD_RuleBase::IsVariable(const GPL_String & string)
{
	static GPL_RegularExpression regexp("\\$\\([A-Za-z0-9_]+\\)");
	return regexp.Matches(string);
}

GPL_Boolean BLD_RuleBase::IsPattern(const GPL_String & string)
{
	// This is sloppy in that it does not account for escaped
	// meta-characters, but those patterns will evaluate to
	// literal strings inside GPL_RegularExpression, so the
	// net effect is not so deliterious.
	return -1 != string.Find('^') || -1 != string.Find('$') ||
		-1 != string.Find('[') || -1 != string.Find('(') ||
		-1 != string.Find(')') || -1 != string.Find('|') ||
		-1 != string.Find('?') || -1 != string.Find('+') ||
		-1 != string.Find('*') || -1 != string.Find('\\');
}

/***********************************************************************
class BLD_Rule
***********************************************************************/
GPL_Boolean BLD_Rule::Matches(const GPL_String & string,
								GPL_Array<GPL_String> & expressions)
{
	if (string == target)
	{
//WBA_APP->Printf("%s == %s", string.GetData(), target.GetData());
		expressions.Resize(1);
		expressions.SetElement(string, 0);
		return TRUE;
	}
//WBA_APP->Printf("%s != %s", string.GetData(), target.GetData());
	return FALSE;
}

/***********************************************************************
class BLD_VariableRule
***********************************************************************/
GPL_Boolean BLD_VariableRule::Matches(const GPL_String & string,
										GPL_Array<GPL_String> & expressions)
{
	const GPL_String t = BLD_CONTEXT.Evaluate(target);
	if (string == t)
	{
//WBA_APP->Printf("%s == %s", string.GetData(), t.GetData());
		expressions.Resize(1);
		expressions.SetElement(string, 0);
		return TRUE;
	}
//WBA_APP->Printf("%s != %s", string.GetData(), t.GetData());
	return FALSE;
}

/***********************************************************************
class BLD_PatternRule
***********************************************************************/
GPL_Boolean BLD_PatternRule::Matches(const GPL_String & string,
										GPL_Array<GPL_String> & expressions)
{
	if (regexp.Matches(string))
	{
//WBA_APP->Printf("%s == %s", string.GetData(), regexp.GetPattern().GetData());
		const unsigned long n = regexp.GetSubExpressionCount();
		expressions.Resize(n);
		for (unsigned long i = 0; n > i; ++i)
		{
			expressions[i] = regexp.GetMatch(i);
		}
		return TRUE;
	}
//WBA_APP->Printf("%s != %s", string.GetData(), regexp.GetPattern().GetData());
	return FALSE;
}

/***********************************************************************
class BLD_VariablePatternRule
***********************************************************************/
GPL_Boolean BLD_VariablePatternRule::Matches(const GPL_String & string,
											GPL_Array<GPL_String> & expressions)
{
	const GPL_String p = BLD_CONTEXT.Evaluate(pattern);
	regexp.SetPattern(p);
	if (regexp.Matches(string))
	{
//WBA_APP->Printf("%s == %s", string.GetData(), regexp.GetPattern().GetData());
		const unsigned long n = regexp.GetSubExpressionCount();
		expressions.Resize(n);
		for (unsigned long i = 0; n > i; ++i)
		{
			expressions[i] = regexp.GetMatch(i);
		}
		return TRUE;
	}
//WBA_APP->Printf("%s != %s", string.GetData(), regexp.GetPattern().GetData());
	return FALSE;
}
